// mpicc a+b2.c -Wall -o test.x

// processor i  - in a ring
// output_i = local_i^2 + local_{i-1}


// exercises

// local_i = local_i^2 + local_{i-1}

// output_i = local_i + local_{i-1}^2

// output_i = local_{i+1} + local_i^2 + local_{i-1}^2



#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <time.h>
#include <math.h>

int main( int argc, char *argv[] )
{
   int i, size=6, me, nProcs;
   MPI_Status status;
   MPI_Request requestS, requestR;

   MPI_Init( &argc, &argv );
   MPI_Comm_rank( MPI_COMM_WORLD, &me );
   MPI_Comm_size( MPI_COMM_WORLD, &nProcs );

   if( me==0 && argc < 2 ){ printf("argument please!\n");  return(-1);}
   size = atoi(argv[1]); 

   int *LocalBuffer, *RecvBuffer, *output;
   LocalBuffer = (int *) malloc( size * sizeof(int) );
   RecvBuffer  = (int *) malloc( size * sizeof(int) );
   output      = (int *) malloc( size * sizeof(int) );
   srand48( me + (unsigned)time((time_t *)NULL) );
   int base = (int) floor(10 * drand48());
   for( i=0; i<size; i++ ) LocalBuffer[i] = base + i; 

   for( i=0; i<size; i++ ) 
      printf( " [%d]  Local[%d]: %d\n", me, i, LocalBuffer[i] );

   double MPItime;
   MPI_Barrier( MPI_COMM_WORLD );
   //   MPItime = MPI_Wtime();



   
   // Solution #1
   /*
   if( me%2 == 0 ){
      printf( "[%d] Sending to %d\n", me, (me+1)%nProcs );
      MPI_Ssend( LocalBuffer, size, MPI_INT, (me+1)%nProcs, 0, MPI_COMM_WORLD  );
      MPI_Recv ( RecvBuffer,  size, MPI_INT, (me-1+nProcs)%nProcs, 0, MPI_COMM_WORLD, &status );
   };
   
   if( me%2 == 1 ){
      printf( "[%d] Receiving from %d\n", me, (me-1+nProcs)%nProcs );
      MPI_Recv ( RecvBuffer,  size, MPI_INT, (me-1+nProcs)%nProcs, 0, MPI_COMM_WORLD, &status );
      MPI_Ssend( LocalBuffer, size, MPI_INT, (me+1)%nProcs, 0, MPI_COMM_WORLD  );
   } 

   for( i=0; i<size; i++ )
      output[i] = LocalBuffer[i] * LocalBuffer[i];
      //LocalBuffer[i] = LocalBuffer[i] * LocalBuffer[i];
      */




   // Solution #1.5
   /*
   MPI_Sendrecv( LocalBuffer, size, MPI_INT,
                 (me+1)%nProcs, 0,
                 RecvBuffer,  size, MPI_INT,
                 (me-1+nProcs)%nProcs, 0,
                 MPI_COMM_WORLD, &status );

   for( i=0; i<size; i++ )
      output[i] = LocalBuffer[i] * LocalBuffer[i];
   */




   


   
   // Solution #2
   /*
   for( i=0; i<size; i++ )
      output[i] = LocalBuffer[i] * LocalBuffer[i]; 
   MPI_Isend( LocalBuffer, size, MPI_INT, (me+1)%nProcs, 0, MPI_COMM_WORLD, &requestS );
   MPI_Recv ( RecvBuffer,  size, MPI_INT, (me-1+nProcs)%nProcs, 0, MPI_COMM_WORLD, &status );   
   */
   //   vs. 
   // winner!
   /*
   MPI_Isend( LocalBuffer, size, MPI_INT, (me+1)%nProcs, 0, MPI_COMM_WORLD, &requestS );
   for( i=0; i<size; i++ )
      output[i] = LocalBuffer[i] * LocalBuffer[i]; 
   MPI_Recv ( RecvBuffer,  size, MPI_INT, (me-1+nProcs)%nProcs, 0, MPI_COMM_WORLD, &status );   
   */
   //   vs.
   /*
   MPI_Isend( LocalBuffer, size, MPI_INT, (me+1)%nProcs, 0, MPI_COMM_WORLD, &requestS );
   MPI_Recv ( RecvBuffer,  size, MPI_INT, (me-1+nProcs)%nProcs, 0, MPI_COMM_WORLD, &status );   
   for( i=0; i<size; i++ )
      output[i] = LocalBuffer[i] * LocalBuffer[i]; 
   */






   

   // Solution #3
   // good stuff
   
   MPI_Isend( LocalBuffer, size, MPI_INT, (me+1)%nProcs, 0, MPI_COMM_WORLD, &requestS );
   MPI_Irecv( RecvBuffer,  size, MPI_INT, (me-1+nProcs)%nProcs, 0, MPI_COMM_WORLD, &requestR );   
   for( i=0; i<size; i++ )
      output[i] = LocalBuffer[i] * LocalBuffer[i]; 
   MPI_Wait( &requestR, &status );

   // also ok
   /*
   MPI_Irecv( RecvBuffer,  size, MPI_INT, (me-1+nProcs)%nProcs, 0, MPI_COMM_WORLD, &requestR );   
   MPI_Isend( LocalBuffer, size, MPI_INT, (me+1)%nProcs, 0, MPI_COMM_WORLD, &requestS );
   for( i=0; i<size; i++ )
      output[i] = LocalBuffer[i] * LocalBuffer[i]; 
   MPI_Wait( &requestR, &status );
   */

   // dreadful
   /*
   MPI_Isend( LocalBuffer, size, MPI_INT, (me+1)%nProcs, 0, MPI_COMM_WORLD, &requestS );
   MPI_Irecv( RecvBuffer,  size, MPI_INT, (me-1+nProcs)%nProcs, 0, MPI_COMM_WORLD, &requestR );   
   MPI_Wait( &requestR, &status );
   for( i=0; i<size; i++ )
      output[i] = LocalBuffer[i] * LocalBuffer[i]; 
   */

   
   // what's the point?  == Recv
   /*
   MPI_Isend( LocalBuffer, size, MPI_INT, (me+1)%nProcs, 0, MPI_COMM_WORLD, &requestS );
   for( i=0; i<size; i++ )
      output[i] = LocalBuffer[i] * LocalBuffer[i]; 
   MPI_Irecv( RecvBuffer,  size, MPI_INT, (me-1+nProcs)%nProcs, 0, MPI_COMM_WORLD, &requestR );   
   MPI_Wait( &requestR, &status );
   */

   
   
   for( i=0; i<size; i++ ) 
      output[i] += RecvBuffer[i]; 

   for( i=0; i<size; i++ ) 
      printf( " [%d]  \t\tResult[%d]: %d\n", me, i, output[i] );

   // MPI_Barrier( MPI_COMM_WORLD );
   // MPItime = MPI_Wtime() - MPItime;
   // printf( " [%d]  \t\ttime: %lf\n", me, MPItime );
      
   free( LocalBuffer );
   free( RecvBuffer );
   free( output );
   
   MPI_Finalize();
   return 0;
}
